import altair


LIGHT_GRAY = '#d3d3d3'
DARK_GRAY = '#a9a9a9'


def conjurer():
    return dict(
        config=dict(
            axis=dict(
                grid=True,
                gridColor=LIGHT_GRAY,
                tickColor=LIGHT_GRAY,
                domainColor=LIGHT_GRAY,
                labelColor=DARK_GRAY,
                titleColor=DARK_GRAY
            ),
            legend=dict(
                orient='top',
                titleAnchor='start',
                columns=0,
                titleColor=DARK_GRAY,
                labelColor=DARK_GRAY
            ),
            title=dict(
                align='left',
                anchor='start'
            )
        ),
    )


def apply_theme():
    altair.themes.register('conjurer', conjurer)
    altair.themes.enable('conjurer')



from sklearn import (
    pipeline,
    impute,
    preprocessing
)


def add_default_preprocessing(name, ml_estimator):
    return pipeline.Pipeline(
        steps=[
            ("mvi", impute.SimpleImputer()),
            ("std", preprocessing.StandardScaler()),
            (name, ml_estimator)
        ]
    )


def convert_param_dict(name, params_dict):
    return {
        "{}__{}".format(name, key): item
        for key, item in params_dict.items()
    }



"""
Functions to be used for quick data check / exploration
"""

from .api import *

__all__ = [
    "check_stats",
    "check_series",
    "get_unique_values",
    "get_columns_in_dfs",
    "get_fk_coverage"
]



import altair as alt

from conjurer.logic.eda.vis import binning


altair_max_rows = 5000


def plot_scatter(df, column_x, column_y, **kwargs):
    if len(df) > altair_max_rows:
        return plot_heatmap(df, column_x, column_y, **kwargs)
    else:
        return plot_points(df, column_x, column_y, **kwargs)


def plot_heatmap(df, column_x, column_y, num_bins_x=50, num_bins_y=50, xmin=None, xmax=None, ymin=None, ymax=None):
    is_quantitative_x = binning.is_quantitative(df[column_x], num_bins_x)
    is_quantitative_y = binning.is_quantitative(df[column_y], num_bins_y)
    ft_df = binning.create_frequency_table_2d(
        df, column_x, column_y, num_bins_x, num_bins_y, xmin, xmax, ymin, ymax)
    ft_df = ft_df[ft_df[binning.FREQUENCY_CNAME] > 0]
    encode_args = _get_encode_args(column_x, column_y, is_quantitative_x, is_quantitative_y)
    title = "{} vs {} (Heatmap)".format(column_y, column_x)
    chart = alt.Chart(ft_df).encode(**encode_args).properties(title=title)\
        .configure_legend(orient="right", columns=1).interactive()
    return chart.mark_rect() if is_quantitative_x or is_quantitative_y else chart.mark_circle()


def _get_encode_args(column_x, column_y, is_quantitative_x, is_quantitative_y):
    x_tooltip = ["{}_lb".format(column_x), "{}_ub".format(column_x)] if is_quantitative_x else [column_x]
    y_tooltip = ["{}_lb".format(column_y), "{}_ub".format(column_y)] if is_quantitative_y else [column_y]
    common_args = dict(
        size=binning.FREQUENCY_CNAME,
        color=alt.Color(binning.FREQUENCY_CNAME, scale=alt.Scale(scheme="greys")),
        tooltip=x_tooltip + y_tooltip + [binning.FREQUENCY_CNAME]
    )
    x_args = dict(
        x=alt.X("{}_lb".format(column_x), title=column_x),
        x2="{}_ub".format(column_x)
    ) if is_quantitative_x else dict(x=column_x)
    y_args = dict(
        y=alt.X("{}_lb".format(column_y), title=column_y),
        y2="{}_ub".format(column_y)
    ) if is_quantitative_y else dict(y=column_y)
    return {**common_args, **x_args, **y_args}


def plot_points(df, column_x, column_y, xmin=None, xmax=None, ymin=None, ymax=None):
    title = "{} vs {} (Scatter)".format(column_y, column_x)
    return alt.Chart(df).mark_circle().encode(
        x=alt.X(column_x, **_get_axis_args(df, column_x, xmin, xmax)),
        y=alt.Y(column_y, **_get_axis_args(df, column_y, ymin, ymax)),
        tooltip=[column_x, column_y]
    ).properties(title=title).interactive()


def _get_axis_args(df, column, minv, maxv):
    if minv is not None or maxv is not None:
        minv = minv if minv is not None else df[column].min()
        maxv = maxv if maxv is not None else df[column].max()
        return dict(domain=[minv, maxv])
    else:
        return {}



class TunerConfig(object):
    def __init__(self, estimator_dict, distributions_rg, distributions_cl, grid_rg, grid_cl):
        self.estimator_dict = estimator_dict
        self.parameters_dict = self.get_default_parameters(
            distributions_rg, distributions_cl, grid_rg, grid_cl)

    @staticmethod
    def get_default_parameters(distributions, distributions_cl, grid, grid_cl):
        return {
            "grid": {
                **{ptype: grid_cl for ptype in ["cl", "mcl"]},
                **{"rg": grid}
            },
            "random": {
                **{ptype: distributions_cl for ptype in ["cl", "mcl"]},
                **{"rg": distributions}
            }
        }



from scipy import stats
from sklearn import ensemble

from . import base


estimator = {
    "cl": ensemble.RandomForestClassifier(),
    "rg": ensemble.RandomForestRegressor(),
    "mcl": ensemble.RandomForestClassifier()
}
grid = dict(
    max_depth=[2, 4, 10],
    n_estimators=[100, 200],
    min_samples_leaf=[0.01, 0.05]
)
distributions = dict(
    max_depth=list(range(2, 11)),
    n_estimators=[100, 200],
    min_samples_leaf=stats.uniform(0.01, 0.1)
)
config = base.TunerConfig(estimator, distributions, distributions, grid, grid)



from scipy import stats

from . import base
import gbm_autosplit


ESTIMATOR_PARAMS = dict(
    max_depth=8,
    max_n_estimators=5000,
    early_stopping_rounds=100,
    importance_type="gain"
)
estimator = {
    "cl": gbm_autosplit.XGBClassifier(eval_metric="auc", **ESTIMATOR_PARAMS),
    "rg": gbm_autosplit.XGBRegressor(eval_metric="rmse", **ESTIMATOR_PARAMS),
    "mcl": gbm_autosplit.XGBClassifier(eval_metric="accuracy", **ESTIMATOR_PARAMS)
}

grid = dict(
    reg_lambda=[0.0, 0.01, 1.0],
    learning_rate=[0.01, 0.02, 0.1],
    ratio_min_child_weight=[0.0, 0.01, 0.1]
)
distributions = dict(
    reg_lambda=stats.uniform(0.0, 1.0),
    learning_rate=stats.uniform(0.01, 0.5),
    ratio_min_child_weight=stats.uniform(0.0, 0.1)
)
config = base.TunerConfig(estimator, distributions, distributions, grid, grid)



import logging

import altair as alt
from pandas.api import types

from conjurer.logic.eda.vis import binning


logger = logging.getLogger(__name__)


def plot_histogram(values, num_bins=50, normalize=False, minv=None, maxv=None):
    try:
        bin_df = binning.create_frequency_table(values, num_bins, minv, maxv)
    except Exception as e:
        raise e
    return plot_frequency_numeric(bin_df, normalize) if len(bin_df.columns) > 3\
        else plot_frequency_category(bin_df, normalize)


def plot_histogram_for_stats(df, stat_df, num_bins=50, normalize=False):
    for ind, column in enumerate(df.columns):
        minv, maxv = \
            stat_df[stat_df["column_name"] == column][
                ["min", "max"]].values.tolist()[0]
        try:
            plot_histogram(df[column], num_bins, normalize, minv, maxv).display()
        except binning.BinCreationError as e:
            logger.info("Histogram for {} was skipped: {}".format(column, e.message))
            pass


def plot_frequency_numeric(df, normalize, xname=None):
    column_lb = df.columns[0]
    column_ub = df.columns[1]
    xname = xname or column_lb.replace("_lb", "")
    column_y = binning.RATIO_CNAME if normalize else binning.FREQUENCY_CNAME
    y_args = dict(axis=alt.Axis(format="%")) if normalize else {}
    return alt.Chart(df).mark_bar().encode(
        alt.X(column_lb, bin="binned", axis=alt.Axis(title=xname)),
        x2=column_ub,
        y=alt.Y("{}:Q".format(column_y), **y_args),
        tooltip=[column_lb, column_ub, column_y]
    ).properties(height=200, width=800, title="Histogram of {}".format(xname)).interactive()


def plot_frequency_category(df, normalize, xname=None):
    x_args = {} if types.is_integer_dtype(df.dtypes[0]) else dict(sort="-y")
    y_args = dict(axis=alt.Axis(format="%")) if normalize else {}
    column_y = binning.RATIO_CNAME if normalize else binning.FREQUENCY_CNAME
    return alt.Chart(df).mark_bar().encode(
        x=alt.X("{}:N".format(df.columns[0]), **x_args),
        y=alt.Y("{}:Q".format(column_y), **y_args)
    ).properties(height=200, width=800, title="Histogram of {}".format(df.columns[0])).interactive()



import pandas
from pandas.api import types
from IPython.display import display

from conjurer.logic.eda.check import stat_calculator
from conjurer.logic.eda.vis import histogram


def check_stats(df, skip_histogram=False):
    print("[table-wise confirmation]")
    print("shape: {}x{}".format(len(df), len(df.columns)))
    print("duplication: {}".format(len(df[df.duplicated()])))
    print("[column-wise confirmation]")
    stat_df = stat_calculator.calc_column_stat(df)
    display(stat_df)
    if not skip_histogram:
        print("[histogram]")
        histogram.plot_histogram_for_stats(df, stat_df)
    print("[head]")
    display(df.head())
    return stat_df


def check_series(df, unit_keys):
    grb_obj = df.groupby(unit_keys)
    args_dict = {"count": grb_obj.count().apply(max, axis=1)}
    for c in df.columns:
        if c in unit_keys:
            continue
        cnames, values = stat_calculator.calculate_series_stats(c, df[c].dtype, grb_obj)
        for c, v in zip(cnames, values):
            args_dict[c] = v
    return pandas.DataFrame(args_dict)


def get_unique_values(df, columns):
    return stat_calculator.get_unique_values(df, columns)


def get_columns_in_dfs(df_list, name_list):
    return pandas.concat([
        pandas.DataFrame({
            "table_name": [name] * len(df.columns),
            "column_name": df.columns,
            "dtype": [df.dtypes[c] for c in df.columns]
        })
        for df, name in zip(df_list, name_list)
    ], ignore_index=True)


def get_fk_coverage(fk_df, k_df, fk_columns, k_columns, do_print=True):
    """
    Check how many keys in the first df exist in the second df
    Parameters
    ----------
    fk_df : pandas.DataFrame
    k_df : pandas.DataFrame
    fk_columns : str or list of str
    k_columns : str or list of str
    do_print : bool

    Returns
    -------
    float

    """
    keys_in_fk_df = get_unique_values(fk_df, fk_columns)
    keys_in_k_df = get_unique_values(k_df, k_columns)
    num_intersection = len(keys_in_fk_df.intersection(keys_in_k_df))
    num_target_keys = len(keys_in_fk_df)
    if do_print:
        print("{:.2%} ({} / {})".format(num_intersection / num_target_keys, num_intersection, num_target_keys))
    return num_intersection / num_target_keys
